#include "Desktop.h"
#include "Config.h"
#include <cmath>
#include <iostream>
#include "Server.h"
#include <sstream>
#include "Colour.h"

// rofl.
static const char * NUMBERS[] = {
	"zero",
	"one",
	"two",
	"three",
	"four",
	"five",
	"six",
	"seven",
	"eight",
	"nine",
	"ten",
	"eleven",
	"twelve",
	"thirteen",
	"fourteen",
	"fifteen",
	"sixteen",
	"seventeen",
	"eighteen",
	"nineteen",
	"twenty"
};

namespace Visibility {

std::vector<std::string> Desktop::desktop_names;

// let's hope sigc++ will call this before other things attached to the
// signal
// Server::signal_desktop_names_changed.connect(sigc::ptr_fun(&Desktop::update_desktop_names));

static std::string num_to_string(int number) {
	std::stringstream stream;
	stream << number;
	return stream.str();
}

void Desktop::update_desktop_names(void) {
	desktop_names.clear();
	Server * server = Server::instance();
	server->get_desktop_names(desktop_names);
	unsigned num_desktop_names = desktop_names.size();
	int delta = server->get_number_of_desktops() - num_desktop_names;

	unsigned pos = num_desktop_names + 1;
	while (delta > 0) {
		if (pos > 20) {
			desktop_names.push_back(num_to_string(pos));
		} else {
			desktop_names.push_back(NUMBERS[pos]);
		}
		pos++;
		delta--;
	}
}

Desktop::Desktop(::Window parent, CARD32 number) {
	is_active = false;
	this->number = number;
	Config * config = Config::instance();
	main = Window::get_window(0, 0, config->image_size,
		config->image_size + 2*config->spacing, parent);

	// connect signals
	signal_new_window_connection = Window::signal_new.connect(sigc::mem_fun(this, &Desktop::window_opened_callback));
	signal_desktop_changed_connection = Server::signal_desktop_changed.connect(sigc::mem_fun(this, &Desktop::desktop_changed_callback));
	signal_config_changed_connection = config->signal_changed.connect(sigc::mem_fun(this, &Desktop::reconfigure));
	signal_desktop_name_changed_connection = Server::signal_desktop_names_changed.connect(sigc::mem_fun(this, &Desktop::desktop_name_changed_callback));

	initialize();
	main->map();
}

void Desktop::desktop_name_changed_callback(void) {
	name_string = desktop_names[number];
	Config * config = Config::instance();
	if (config->show_desktop_names) {
		this->name.reset(new DesktopName(main->drawable, name_string));
		this->name->window->signal_clicked.connect(sigc::mem_fun(this, &Desktop::desktop_name_clicked_callback));
		this->name->set_active(is_active);
	}
	resize();
	position_internals();
}

void Desktop::initialize(void) {
	Config * config = Config::instance();

	name_string = desktop_names[number];
	if (config->show_desktop_names) {
		this->name.reset(new DesktopName(main->drawable, name_string));
		this->name->window->signal_clicked.connect(sigc::mem_fun(this, &Desktop::desktop_name_clicked_callback));
	}

	// populate with windows
	std::vector< ::Window > windows;
	Server::instance()->get_client_list(windows);
	std::vector< ::Window >::iterator iter;
	for (iter = windows.begin(); iter != windows.end(); iter++) {
		window_opened_callback(*iter);
	}

	resize();
	position_internals();
	main->set_background_colour(config->inactive_bg_colour);
}

void Desktop::resize(void) {
	Config * config = Config::instance();
	width = calculate_width();
	main->resize(width, config->image_size + 2*config->spacing);
	main->set_background_colour(is_active ? config->active_bg_colour : config->inactive_bg_colour);
	main->clear();

	signal_width_changed.emit();
}

void Desktop::desktop_changed_callback(void) {
	is_active = (Server::instance()->get_current_desktop() == number);
	Config * config = Config::instance();
	main->set_background_colour(is_active ? config->active_bg_colour : config->inactive_bg_colour);
	if (config->show_desktop_names) {
		name->set_active(is_active);
	}
	main->clear();
}

unsigned Desktop::calculate_width(void) {
	Config * config = Config::instance();
	unsigned num_icons = window_icons.size();
	unsigned width = num_icons * config->image_size + (num_icons > 1 ? (num_icons - 1) * config->spacing : 0) + (num_icons ? config->text_spacing : 0);
	if (config->show_desktop_names) {
		unsigned name_width = 0, name_height = 0; // height is unused
		name->get_size(name_width, name_height);
		width += name_width;
	}

	if (!width) width = config->image_size;

	return width;
}

void Desktop::position_internals(void) {
	Config * config = Config::instance();

	// position text (desktop name)

	unsigned offset = 0;

	if (config->show_desktop_names) {
		unsigned name_width = 0, name_height = 0;
		name->get_size(name_width, name_height);
		name->window->move(0, unsigned((config->image_size + 2*config->spacing - name_height) / 2.0));
		offset += name_width + (window_icons.size() ? config->text_spacing : 0);
		name->window->clear();
	}

	// windows

	std::map< ::Window, std::tr1::shared_ptr<WindowIcon> >::iterator iter;
	for (iter = window_icons.begin(); iter != window_icons.end(); iter++) {
		iter->second->window->move(offset, config->spacing);
		iter->second->window->map();
		offset += config->spacing + config->image_size;
	}
}

void Desktop::window_opened_callback(::Window window) {
	std::tr1::shared_ptr<Window> new_win = Window::get_window(window);

	// sanity checks

	if (new_win->get_skip_pager()) return;
	CARD32 win_desktop = new_win->get_desktop();
	if ((win_desktop != number) && (win_desktop != 0xFFFFFFFF)) return;

	new_win->select_input(PropertyChangeMask);
	window_signals[window].push_back(new_win->signal_closed.connect(sigc::mem_fun(this, &Desktop::window_closed_callback)));
	std::tr1::shared_ptr<WindowIcon> window_icon(new WindowIcon(main->drawable, window));
	window_icons[window] = window_icon;
	resize();
	position_internals();
}

void Desktop::window_closed_callback(::Window window) {
	if (window_icons.find(window) != window_icons.end()) {
		std::tr1::shared_ptr<WindowIcon> window_icon = window_icons[window];

		// disconnect signals first!
		std::vector<sigc::connection>::iterator iter;
		for (iter=window_signals[window].begin();iter!=window_signals[window].end();iter++) {
			iter->disconnect();
		}

		window_signals[window].clear();
		window_signals.erase(window);

		window_icons.erase(window);
		resize();
		position_internals();
	}
}

void Desktop::reconfigure(void) {
	if (!Config::instance()->show_desktop_names) {
		name.reset();
	} else {
		if (!name) {
			name.reset(new DesktopName(main->drawable, name_string));
			this->name->window->signal_clicked.connect(sigc::mem_fun(this, &Desktop::desktop_name_clicked_callback));
			name->set_active(is_active);
		} else {
			name->reconfigure();
		}
	}
	resize();
	position_internals();
}

void Desktop::desktop_name_clicked_callback(::Window window) {
	Server::instance()->set_current_desktop(number);
}

Desktop::~Desktop() {
	signal_new_window_connection.disconnect();
	signal_desktop_changed_connection.disconnect();
	signal_config_changed_connection.disconnect();
	signal_desktop_name_changed_connection.disconnect();

	main->unmap();
	main->destroy();
}

}
