#include "MainLoop.h"
#include <iostream>

#include <memory>

namespace MainLoop {

	std::vector< std::tr1::shared_ptr<Timeout> > MainLoop::timeouts;
	std::vector< std::tr1::shared_ptr<FunctorBase> > MainLoop::idles;
	std::map< int, std::tr1::shared_ptr< Functor2Base<int> > > MainLoop::file_descriptors;

	Timeout::Timeout(std::tr1::shared_ptr<FunctorBase> functor, unsigned interval) {
		interval_ = interval;
		functor_ = functor;

		next_call_time = Time() + Time(0, interval_);
	}

	Timeout::~Timeout() {
	}

	bool Timeout::tick(void) {
		Time time;
		bool retval = true;

		if (next_call_time <= time) {
			retval = (*functor_)();
			// update next call time
			next_call_time += Time(0, interval_);
		}

		return retval;
	}

	PtrFunctor::PtrFunctor(bool (*function)()) {
		function_ = function;
	}

	bool PtrFunctor::operator()(void) const {
		if (disconnect_) return false;
		return (*function_)();
	}

	// converts a vector of ints to an fd set. returns the highest numbered fd
	// plus 1.
	static int vector_to_fd_set(std::vector<int> file_descriptors,
															fd_set * set) {

		int retval = 0;
		FD_ZERO(set);
		std::vector<int>::const_iterator iter;
		for (iter = file_descriptors.begin(); iter != file_descriptors.end(); iter++) {
			if (*iter > retval) retval = *iter;
			FD_SET(*iter, set);
		}
		return ++retval;
	}

	// returns a vector of the keys of 'map'
	template<class key_type, class value_type>
	static void keys(std::map<key_type, value_type> &map, std::vector<key_type> &keys) {
		typename std::map<key_type, value_type>::const_iterator iter;

		for (iter = map.begin(); iter != map.end(); iter++) {
			keys.push_back(iter->first);
		}
	}

	std::tr1::shared_ptr<PtrFunctor> ptr_fun(bool (*function)()) {
		return std::tr1::shared_ptr<PtrFunctor>(new PtrFunctor(function));
	}

	int MainLoop::get_fd_set(fd_set * set) {
		int nfds = 0;
		if (file_descriptors.size()) {
			std::vector<int> fds;
			fds.reserve(file_descriptors.size());
			keys(file_descriptors, fds);
			nfds = vector_to_fd_set(fds, set);
		}
		return nfds;
	}

	void MainLoop::call_idle_functions(void) {
		std::vector< std::tr1::shared_ptr<FunctorBase> >::iterator iter = idles.begin();

		while (iter != idles.end()) {
			if (!(**iter)()) {
				// remove this idle func; it returned false.
				idles.erase(iter);
				iter = idles.begin(); // erase invalidates iters
			} else {
				iter++;
			}
		}
	}

	void MainLoop::tick_timeout_functions(Time & timeout) {
		if (timeouts.size()) {
			// tick the timeouts
			timeout = (*timeouts.begin())->next_call_time - Time();
			std::vector< std::tr1::shared_ptr<Timeout> >::iterator iter = timeouts.begin();
			while (iter != timeouts.end()) {
				if (!(*iter)->tick()) {
					// remove this one; this means we have to start all over too,
					// because 'erase' invalidates iters.
					// delete *iter;
					timeouts.erase(iter);
					iter = timeouts.begin();
				} else {
					Time new_time = (*iter)->next_call_time - Time();
					if (new_time < timeout) {
						timeout = new_time;
					}
					iter++;
				}
			}
		} else {
			// have select() wait half a second if there are no timeouts.
			timeout.set(0, (unsigned)(0.5e6));
		}
	}

	void MainLoop::run(void) {
		while (1) {
			fd_set set;
			int nfds = 0;
			int select_return = -1;

			// wait until it is time for a timeout to occur
			struct timeval timeval;

			Time timeout(0, 0);
			if (idles.size()) {
				do {
					// we have time to spare. check the file descriptors; if there is no
					// data available, call the idle functions.
					if (file_descriptors.size()) {
						Time(0, 0).get(timeval); // used to tell select() not to wait
						nfds = get_fd_set(&set);
						select_return = select(nfds, &set, 0, 0, &timeval);
						// nothing interesting, etc.
						if (select_return <= 0) call_idle_functions();
					} else {
						call_idle_functions();
					}
					tick_timeout_functions(timeout);
				} while ((timeout >= Time(0, 500)) && (select_return <= 0));
			} else if (timeouts.size()) {
				tick_timeout_functions(timeout);
			} else {
				timeout.set(0, (unsigned)(0.5e6));
			}

			timeout.get(timeval);

			// prepare the fd set
			nfds = get_fd_set(&set);
			select_return = select(nfds, nfds? &set : 0, NULL, NULL, &timeval);
			if (select_return <= 0) continue; // nothing interesting in the fds
			// something interesting happened.
			std::map< int, std::tr1::shared_ptr< Functor2Base<int> > >::const_iterator iter;

			// call the callback functions for each fd that is good to go
			for (iter = file_descriptors.begin(); iter != file_descriptors.end(); iter++) {
				FD_ZERO(&set);
				FD_SET(iter->first, &set);
				// set timeval to 0
				Time(0, 0).get(timeval);
				select_return = select(iter->first + 1, &set, NULL, NULL, &timeval);
				if (select_return <= 0) continue;
				// this one is good to go!
				(*iter->second)(iter->first);
			}
		}
	}

}
