#include "Pager.h"
#include "Server.h"
#include "Config.h"
#include <iostream>
#include <cmath>
#include <sstream>

namespace Visibility {

	Pager::Pager(void) {
		this->number_of_desktops = Server::instance()->get_number_of_desktops();
		initialize();
		Server::signal_number_of_desktops_changed.connect(sigc::mem_fun(this, &Pager::number_of_desktops_changed_callback));
		Window::signal_window_desktop_changed.connect(sigc::mem_fun(this, &Pager::window_changed_desktop_callback));
		Config::instance()->signal_changed.connect(sigc::mem_fun(this, &Pager::reinitialize));
		Server::signal_desktop_changed.connect(sigc::mem_fun(this, &Pager::current_desktop_changed_callback));
	}

	void Pager::fill_pixmap(std::tr1::shared_ptr<Pixmap> pixmap) {
		Config * config = Config::instance();
		pixmap->fill_rectangle(config->border_colour, 0, 0, width, height);
		pixmap->fill_rectangle(config->bg_colour, config->border_width, config->border_width, width - 2*config->border_width, height - 2*config->border_width);
	}

	void Pager::position_internals(void) {	// position the desktops correctly
		Config * config = Config::instance();

		unsigned offset = config->border_width + config->spacing;

		if (config->single_desktop_mode) {
			desktops[current_desktop]->main->move(offset, config->border_width);
			desktops[current_desktop]->main->map();
		}

		std::vector< std::tr1::shared_ptr<Desktop> >::iterator iter;
		for (iter = desktops.begin(); iter != desktops.end(); iter++) {
			if (config->single_desktop_mode && unsigned(iter - desktops.begin()) != current_desktop) {
				(*iter)->main->unmap();
			} else {
				(*iter)->main->move(offset, config->border_width);
				(*iter)->main->map();
				offset += config->desktop_separation + (*iter)->get_width();
			}
		}
	}

	void Pager::position(void) {
		Config * config = Config::instance();

		unsigned x = 0, y = 0;
		unsigned root_width = 0, root_height = 0;

		Server::instance()->get_display_size(root_width, root_height);

		if (config->orientation == "bottom_right" || config->orientation == "top_right") {
			// keep right side fixed
			x = root_width - (config->gap_x + width);
		} else {
			// keep the left side fixed
			x = config->gap_x;
		}
		if (config->orientation == "bottom_right" || config->orientation == "bottom_left") {
			// keep the bottom fixed
			y = root_height - (config->gap_y + height);
		} else {
			// keep the top fixed
			y = config->gap_y;
		}

		main->move(x, y);

		if (config->set_partial_strut) {
			// set strut
			CARD32 size = height + config->gap_y;
			Window::StrutType strut_type;
			if (config->orientation.find("top") == 0) {
				// set top strut
				strut_type = Window::STRUT_TOP;
			} else {
				// set bottom strut
				strut_type = Window::STRUT_BOTTOM;
			}
			main->set_partial_strut(strut_type, size, x, x + width);
		} else {
			main->remove_partial_strut();
		}
	}

	unsigned Pager::calculate_width(void) {
		Config * config = Config::instance();

		unsigned ret_val = 2 * (config->border_width + config->spacing);
		if (config->single_desktop_mode) {
			ret_val += desktops[current_desktop]->get_width();
			return ret_val;
		}

		std::vector< std::tr1::shared_ptr<Desktop> >::const_iterator iter;
		for (iter = desktops.begin(); iter != desktops.end(); iter++) {
			ret_val += (*iter)->get_width();
		}

		ret_val += (desktops.size() - 1) * config->desktop_separation;

		return ret_val;
	}

	void Pager::initialize(void) {
		Server * server = Server::instance();
		main = Window::get_window(0, 0, 500, 500);
		// get expose events for the main window
		main->select_input(ExposureMask | PropertyChangeMask);
		main->set_decorated(false);
		main->set_state(Window::State(Window::STICKY | Window::SKIP_PAGER | Window::SKIP_TASKBAR), true);

		if (Config::instance()->set_window_type) {
			main->set_window_type(Window::WINDOW_TYPE_DOCK);
		}

		current_desktop = server->get_current_desktop();
		active_window = Window::get_active_window();

		hints = XAllocSizeHints ();
		hints->flags = PPosition | PSize;
		XSetWMNormalHints (server->display, main->drawable, hints);

		wm_hints = XAllocWMHints();
		wm_hints->input = False;
		wm_hints->flags = InputHint;
		XSetWMHints(server->display, main->drawable, wm_hints);

		// set up desktops
		for (unsigned count = 0; count < number_of_desktops; count++) {
			std::tr1::shared_ptr<Desktop> desktop(new Desktop(main->drawable, count));
			desktops.push_back(desktop);
			desktop->signal_width_changed.connect(sigc::mem_fun(this, &Pager::resize));
		}

		main->set_desktop(0xFFFFFFFF);
		resize();

		main->map();
		server->flush();

		Server::signal_desktop_changed.emit();
		Window::signal_active_window_changed.emit();

	}

	Pager::~Pager() {
		main->unmap();
		main->destroy();
		XFree(hints);
		XFree(wm_hints);
	}

	void Pager::resize(void) {
		Config * config = Config::instance();

		// modify size hints first
		hints->flags = PPosition | PSize | PMinSize | PMaxSize;
		hints->min_width = 0;
		hints->max_width = 65535;
		hints->min_height = 0;
		hints->max_height = 65535;
		Server * server = Server::instance();
		XSetWMNormalHints (server->display, main->drawable, hints);

		width = calculate_width();
		height = (config->border_width + config->spacing) * 2 + config->image_size;
		main->resize(width, height);

		//std::cout << width << "x" << height << std::endl;

		hints->flags = PPosition | PSize | PMinSize | PMaxSize;
		hints->min_width = hints->max_width = width;
		hints->min_height = hints->max_height = height;
		XSetWMNormalHints (server->display, main->drawable, hints);

		std::tr1::shared_ptr<Pixmap> pixmap(new Pixmap(main->drawable, width, height));
		fill_pixmap(pixmap);
		main->set_background_pixmap(pixmap);
		main->clear();
		position();
		position_internals();
		main->set_desktop(0xFFFFFFFF);
		XSetWMHints(server->display, main->drawable, wm_hints);
		Server::instance()->flush();
	}

	// TODO: this really needs to be optimized
	void Pager::window_changed_desktop_callback(::Window window) {
		if (window == main->drawable) {
			// someone seems to have moved us!
			// we must be moved back.
			main->set_desktop(0xFFFFFFFF);
			return;
		}

		// tricks.

		std::tr1::shared_ptr<Window> changed = Window::get_window(window);
		changed->signal_closed.emit(window);
		Window::signal_new.emit(window);
		Window::signal_active_window_changed.emit();
	}

	void Pager::reinitialize(void) {
		Tooltip::height = 0;

		resize();
	}

	void Pager::number_of_desktops_changed_callback(void) {
		number_of_desktops = Server::instance()->get_number_of_desktops();
		int count = desktops.size() - number_of_desktops;
		if (count < 0) {
			// desktops were added
			while (count < 0) {
				std::tr1::shared_ptr<Desktop> desktop(new Desktop(main->drawable,
																													number_of_desktops + count));
				desktops.push_back(desktop);
				desktop->signal_width_changed.connect(sigc::mem_fun(this, &Pager::resize));
				count++;
			}
		} else {
			// desktops were removed
			while (count--) {
				desktops.pop_back();
			}
		}
		resize();
	}

	void Pager::current_desktop_changed_callback(void) {
		current_desktop = Server::instance()->get_current_desktop();
		if (Config::instance()->single_desktop_mode) {
			resize();
			position_internals();
		}
	}
}
