#include "Window.h"
#include "Server.h"
#include <iostream>
#include "Colour.h"

//#include <X11/Xutil.h> // Urgen

namespace Visibility {

	// statics

	sigc::signal<void> Window::signal_active_window_changed;
	sigc::signal<void, ::Window> Window::signal_new;
	sigc::signal<void, ::Window> Window::signal_window_desktop_changed;

	// functions

	std::tr1::shared_ptr<Window> Window::get_window(::Window window) {
		if (windows.find(window) == windows.end()) {
			windows[window].reset(new Window(window));
		}

		return windows[window];
	}

	std::tr1::shared_ptr<Window> Window::get_window(unsigned x, unsigned y, unsigned width, unsigned height, ::Window parent) {
		std::tr1::shared_ptr<Window> window(new Window(x, y, width, height, parent));

		windows[window->drawable] = window;
		return window;
	}

	Window::Window(::Window drawable) {
		this->drawable = drawable;
		signal_closed.connect(sigc::mem_fun(this, &Window::clean_up));
	}

	Window::Window(unsigned x, unsigned y, unsigned width, unsigned height, ::Window parent) {
		Server * server = Server::instance();
		drawable = XCreateSimpleWindow(server->display, parent ? parent : server->root_window, x, y,
																	 width, height, 0, 0, 0);
		signal_closed.connect(sigc::mem_fun(this, &Window::clean_up));
	}

	bool Window::map(void) {
		Server * server = Server::instance();
		if (XMapWindow(server->display, drawable) == Success) {
			return true;
		}
		return false;
	}

	void Window::get_position(unsigned& x, unsigned& y) {
		Server * server = Server::instance();
		::Window bs;

		// let us hope that this is never negative
		XTranslateCoordinates(server->display, drawable, server->root_window, 0, 0,
													(int*)&x, (int*)&y, &bs);
	}

	void Window::unmap(void) {
		Server * server = Server::instance();
		XUnmapWindow(server->display, drawable);
	}


	Window::~Window() {
		//Server * server = Server::instance();
		// scary effects.
		// XDestroyWindow(server->display, drawable);

	}

	std::string Window::get_title(void) {
		std::string string;

		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		unsigned char * property_value;

		int result;

		result = XGetWindowProperty(server->display, drawable, server->wm_name_atom, 0, 0x7fffffff, False, XA_STRING, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, &property_value);
		if (result == Success && property_value) {
			string = std::string((char*)property_value);
			XFree(property_value);
		}

		if (XGetWindowProperty(server->display, drawable, server->net_wm_icon_name_atom, 0, 0x7fffffff, False, server->utf8_string_type, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, &property_value) != Success) {
			return string;
		}

		if (property_value == NULL) {
			return string;
		}

		string = std::string((char*)property_value);

		XFree(property_value);

		return string;
	}

	bool Window::get_class(std::string& name, std::string& class_name) {
		XClassHint * class_hint = XAllocClassHint();
		Server * server = Server::instance();
		XGetClassHint(server->display, drawable, class_hint);
		if (class_hint->res_name != NULL) {
			name = std::string(class_hint->res_name);
			XFree(class_hint->res_name);
		}
		if (class_hint->res_class != NULL) {
			class_name = std::string(class_hint->res_class);
			XFree(class_hint->res_class);
		}
		XFree(class_hint);

		return true;
	}

	void Window::get_icons(std::vector<ARGBData>& icons) {
		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		long * data = NULL;

		if (XGetWindowProperty(server->display, drawable, server->net_wm_icon_atom, 0, 0xffffffff, False, XA_CARDINAL, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, (unsigned char**)(void*)&data) == Success) {
			// determine the number of valid icons
			unsigned num_icons = 0;
			unsigned pos = 0;
			while (pos < nitems_return) {
				unsigned width = data[pos++];
				unsigned height = data[pos++];
				pos += width * height;
				if (pos > nitems_return || width * height == 0) break;
				num_icons++;
			}

			// store the icons
			pos = 0;
			while (num_icons--) {
				unsigned width = data[pos++];
				unsigned height = data[pos++];
				if (pos + width * height > nitems_return) break;
				icons.push_back(ARGBData(&data[pos], width, height));
				pos += width * height;
			}
		}
		if (data) XFree(data);
		data = 0;
		/*if (!icons.size()) {
		// try KWM stuff
		if (XGetWindowProperty(server->display, drawable, server->kwm_icon_atom, 0, 0xffffffff, False, server->kwm_icon_atom, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, (unsigned char**)(void*)&data) == Success) {
		if (nitems_return >= 1) {
		// perhaps success!
		icons.push_back(ARGBData(data[0]));
		}
		}
		}
		if (data) XFree(data);*/
		if (!icons.size()) {
			// no other icons; try XWMHints.
			XWMHints * hints = XGetWMHints(server->display, drawable);
			if (hints) {
				if (hints->flags & IconPixmapHint) {
					try {
						icons.push_back(ARGBData(hints->icon_pixmap, (hints->flags & IconMaskHint ? hints->icon_mask : 0)));
					} catch (...) {} /* fool proof, haha. */
				}
				XFree(hints);
			}
		}
	}

	CARD32 Window::get_desktop(void) {
		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		// not long because long is 64bit on some systems parhaps
		CARD32 * desktop = 0;
		CARD32 ret_val = 0;

		int result;

		result = XGetWindowProperty(server->display, drawable, server->net_wm_desktop_atom, 0, 1, False, XA_CARDINAL, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, (unsigned char**)(void*)&desktop);
		if (result != Success || nitems_return < 1) {
			return 0;
		}
		ret_val = desktop[0];
		XFree(desktop);

		return ret_val;
	}

	bool Window::move(unsigned x, unsigned y) {
		Server * server = Server::instance();
		if (XMoveWindow(server->display, drawable, x, y) == Success) {
			return true;
		}
		return false;
	}

	bool Window::resize(unsigned width, unsigned height) {
		Server * server = Server::instance();
		if (XResizeWindow(server->display, drawable, width, height) == Success) {
			return true;
		}
		return false;
	}

	bool Window::select_input(long event_mask) {
		Server * server = Server::instance();

		if (XSelectInput(server->display, drawable, event_mask) == Success) {
			return true;
		}
		return false;
	}

	bool Window::set_desktop(CARD32 desktop) {
		Server * server = Server::instance();
		XEvent event;
		long mask = SubstructureRedirectMask | SubstructureNotifyMask;

		event.xclient.type = ClientMessage;
		event.xclient.serial = 0;
		event.xclient.send_event = True;
		event.xclient.display = server->display;
		event.xclient.window = drawable;
		event.xclient.message_type = server->net_wm_desktop_atom;

		event.xclient.format = 32;
		event.xclient.data.l[0] = desktop;
		event.xclient.data.l[1] = 0;
		event.xclient.data.l[2] = 0;
		event.xclient.data.l[3] = 0;
		event.xclient.data.l[4] = 0;

		if (XSendEvent(server->display, server->root_window, False, mask, &event)) {
			return true;
		}
		return false;
	}

	bool Window::activate(void) {
		Server * server = Server::instance();
		XEvent event;
		long mask = SubstructureRedirectMask | SubstructureNotifyMask;

		event.xclient.type = ClientMessage;
		event.xclient.serial = 0;
		event.xclient.send_event = True;
		event.xclient.display = server->display;
		event.xclient.window = drawable;
		event.xclient.message_type = server->net_active_window_atom;

		event.xclient.format = 32;
		event.xclient.data.l[0] = 2;
		event.xclient.data.l[1] = 0;
		event.xclient.data.l[2] = 0;
		event.xclient.data.l[3] = 0;
		event.xclient.data.l[4] = 0;

		if (XSendEvent(server->display, server->root_window, False, mask, &event)) {
			return true;
		}
		return false;
	}

	bool Window::set_decorated(bool decorated) {
		// TODO: make the bool work

		Server * server = Server::instance();

		long prop[5] = { 2, 0, 0, 0, 0 };
		int ret;
		ret = XChangeProperty (server->display, drawable,
													 XInternAtom(server->display, "_MOTIF_WM_HINTS", False),
													 XInternAtom(server->display, "_MOTIF_WM_HINTS", False), 32, PropModeReplace,
													 (unsigned char *) prop, 5);
		if (ret == Success) {
			return true;
		}
		return false;
	}

	bool Window::get_skip_pager(void) {
		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		// not long because long is 64bit on some systems parhaps
		Atom * atoms = 0;

		int result;

		result = XGetWindowProperty(server->display, drawable,
																server->net_wm_state_atom, 0, 12, False, XA_ATOM, &actual_type_return,
																&actual_format_return, &nitems_return, &bytes_after_return,
																(unsigned char**)(void*)&atoms);

		if (result != Success) {
			return true;
		}

		// look for the atom we want
		unsigned count = 0;
		for (count = 0; count < nitems_return; count++) {
			if (atoms[count] == server->net_wm_state_skip_pager_atom) {
				XFree(atoms);
				return true;
			}
		}

		// not found
		XFree(atoms);

		result = XGetWindowProperty(server->display, drawable,
																server->net_wm_window_type_atom, 0, 12, False, XA_ATOM, &actual_type_return,
																&actual_format_return, &nitems_return, &bytes_after_return,
																(unsigned char**)(void*)&atoms);

		if (result != Success) {
			return true;
		}

		for (count = 0; count < nitems_return; count++) {
			if (atoms[count] == server->net_wm_window_type_desktop_atom || atoms[count] == server->net_wm_window_type_dock_atom || atoms[count] == server->net_wm_window_type_toolbar_atom || atoms[count] == server->net_wm_window_type_menu_atom || atoms[count] == server->net_wm_window_type_splash_atom) {
				XFree(atoms);
				return true;
			}
		}

		XFree(atoms);

		return false;

	}

	void Window::get_state(std::vector<Atom> &atoms) {
		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		// not long because long is 64bit on some systems parhaps
		Atom * atom_return = 0;

		int result;

		result = XGetWindowProperty(server->display, drawable,
																server->net_wm_state_atom, 0, 12, False, XA_ATOM, &actual_type_return,
																&actual_format_return, &nitems_return, &bytes_after_return,
																(unsigned char**)(void*)&atom_return);

		if (result != Success || nitems_return == 0) return;

		for (unsigned pos = 0; pos < nitems_return; pos++) {
			atoms.push_back(atom_return[pos]);
		}
		XFree(atom_return);
	}

	void Window::get_state(State& state_) {
		Server * server = Server::instance();

		unsigned char state = 0;
		// XWM Hints
		XWMHints * hints = XGetWMHints(server->display, drawable);
		if (hints && (hints->flags & XUrgencyHint)) {
			state |= DEMANDS_ATTENTION;
		}
		XFree(hints);

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		// not long because long is 64bit on some systems parhaps
		Atom * atom_return = 0;

		int result;

		result = XGetWindowProperty(server->display, drawable,
																server->net_wm_state_atom, 0, 12, False, XA_ATOM, &actual_type_return,
																&actual_format_return, &nitems_return, &bytes_after_return,
																(unsigned char**)(void*)&atom_return);

		if (result != Success || nitems_return == 0) {
			if (atom_return) XFree(atom_return);
			state_ = State(state);
			return;
		}

		for (unsigned pos = 0; pos < nitems_return; pos++) {
			if (atom_return[pos] == server->net_wm_state_hidden_atom) {
				state |= HIDDEN;
			} else if (atom_return[pos] == server->net_wm_state_sticky_atom) {
				state |= STICKY;
			} else if (atom_return[pos] == server->net_wm_state_skip_taskbar_atom) {
				state |= SKIP_TASKBAR;
			} else if (atom_return[pos] == server->net_wm_state_skip_pager_atom) {
				state |= SKIP_PAGER;
			} else if (atom_return[pos] == server->net_wm_state_shaded_atom) {
				state |= SHADED;
			} else if (atom_return[pos] == server->net_wm_state_attention_atom) {
				state |= DEMANDS_ATTENTION;
			}
		}

		state_ = State(state);

		XFree(atom_return);

	}

	bool Window::iconify(void) {
		Server * server = Server::instance();
		if (XIconifyWindow(server->display, drawable, server->default_screen) == Success) {
			set_state(HIDDEN);
			return true;
		}
		return false;
	}

	bool Window::set_state(State state, bool replace) {
		Server * server = Server::instance();
		std::vector<Atom> atoms;

		if (!replace) {
			get_state(atoms);
		}

		if (state & STICKY) {
			atoms.push_back(server->net_wm_state_sticky_atom);
		}
		if (state & SKIP_TASKBAR) {
			atoms.push_back(server->net_wm_state_skip_taskbar_atom);
		}
		if (state & SKIP_PAGER) {
			atoms.push_back(server->net_wm_state_skip_pager_atom);
		}
		if (state & HIDDEN) {
			atoms.push_back(server->net_wm_state_hidden_atom);
		}

		int result;
		result = XChangeProperty(server->display, drawable, server->net_wm_state_atom, XA_ATOM, 32, PropModeReplace, (unsigned char*)&(*atoms.begin()), atoms.size());
		if (result == Success) {
			return true;
		}
		return false;
	}

	// TODO: poor implementation
	bool Window::set_window_type(WindowType type) {
		Server * server = Server::instance();
		Atom type_atom;
		switch(type) {
		case WINDOW_TYPE_DOCK:
			type_atom = server->net_wm_window_type_dock_atom;
			break;
		default:
			break;
		}
		int result;
		result = XChangeProperty(server->display, drawable, server->net_wm_window_type_atom, XA_ATOM, 32, PropModeReplace, (unsigned char*)&type_atom, 1);
		if (result == Success) {
			return true;
		}
		return false;
	}

	bool Window::get_frame_extents(CARD32& left, CARD32& right, CARD32& top,
																 CARD32& bottom) {

		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		// not long because long is 64bit on some systems parhaps
		CARD32 * extents = 0;

		int result;

		result = XGetWindowProperty(server->display, drawable,
																server->net_frame_extents_atom, 0, 32, False, XA_CARDINAL, &actual_type_return,
																&actual_format_return, &nitems_return, &bytes_after_return,
																(unsigned char**)(void*)&extents);

		if (result != Success || nitems_return == 0) {
			return false;
		}

		left = extents[0];
		right = extents[1];
		top = extents[2];
		bottom = extents[3];

		XFree(extents);
		return true;
	}

	bool Window::destroy(void) {
		Server * server = Server::instance();
		bool success = false;
		if (XDestroyWindow(server->display, drawable) == Success) {
			success = true;
		}
		return success;
	}

	void Window::clean_up(::Window drawable) {
		windows.erase(drawable); // clear self up
	}

	bool Window::set_background_pixmap(std::tr1::shared_ptr<Pixmap> pixmap) {
		Server * server = Server::instance();
		if (XSetWindowBackgroundPixmap(server->display, drawable, pixmap->drawable) == Success) {
			return true;
		}
		return false;
	}

	bool Window::set_background_colour(CARD32 colour) {
		Server * server = Server::instance();
		colour = Colour(colour).pixel_value(server->depth);
		if (XSetWindowBackground(server->display, drawable, colour) == Success) {
			return true;
		}
		return false;
	}

	bool Window::clear(void) {
		Server * server = Server::instance();
		if (XClearWindow(server->display, drawable) == Success) {
			return true;
		}
		return false;
	}

	bool Window::set_partial_strut(StrutType type, CARD32 size, CARD32 start, CARD32 end) {
		CARD32 data[12] = {0};
		switch(type) {
		case STRUT_LEFT:
			data[0] = size;
			data[4] = start;
			data[5] = end;
			break;
		case STRUT_RIGHT:
			data[1] = size;
			data[6] = start;
			data[7] = end;
			break;
		case STRUT_TOP:
			data[2] = size;
			data[8] = start;
			data[9] = end;
			break;
		case STRUT_BOTTOM:
			data[3] = size;
			data[10] = start;
			data[11] = end;
			break;
		}
		int ret;
		Server * server = Server::instance();
		ret = XChangeProperty(server->display, drawable,
													XInternAtom(server->display, "_NET_WM_STRUT_PARTIAL", False),
													XA_CARDINAL, 32, PropModeReplace, (unsigned char *) data, 12);

		if (ret == Success) return true;
		return false;
	}

	bool Window::remove_partial_strut(void) {
		Server * server = Server::instance();
		int ret;
		ret = XDeleteProperty(server->display, drawable, XInternAtom(server->display, "_NET_WM_STRUT_PARTIAL", False));
		if (ret == Success) return true;
		return false;
	}

	::Window Window::get_active_window(void) {
		Server * server = Server::instance();

		Atom actual_type_return;
		int actual_format_return = 0;
		unsigned long nitems_return = 0;
		unsigned long bytes_after_return = 0;
		::Window * window = 0;
		::Window ret_val = 0;

		int result;

		result = XGetWindowProperty(server->display, server->root_window, server->net_active_window_atom, 0, 1, False, XA_WINDOW, &actual_type_return, &actual_format_return, &nitems_return, &bytes_after_return, (unsigned char**)(void*)&window);
		if (result != Success || nitems_return < 1) {
			return 0;
		}
		ret_val = window[0];
		XFree(window);

		return ret_val;
	}

}
