#include "Pager.h"
#include "Server.h"
#include <unistd.h>
#include <stdlib.h>
#include "Config.h"
#include <signal.h>
#include <iostream>
#include "ArgParser.h"
#include "MainLoop.h"

#if DEBUG
#define DEBUG_P printf
#else
void DEBUG_P(...) { return; }
#endif

int x_error_handler(Display * display, XErrorEvent * e) {
#if DEBUG
	std::cout << "X Error." << std::endl;
#endif
	return 0;
}

static bool contains_window(std::vector<Window> client_list, Window window) {
	std::vector<Window>::iterator iter;
	for (iter = client_list.begin(); iter != client_list.end(); iter++) {
		if (*iter == window) {
			return true;
		}
	}
	return false;
}
// this really shouldn't be in this file !??!

struct WindowDeltaPair {
	Window window;
	bool added;
};

static void client_list_delta(std::vector<Window> old_list, std::vector<Window> new_list, std::vector<WindowDeltaPair>& dest) {
	// returns the difference (should be a single window) between two client lists
	// added will be set to true if the window appears in the new list but not
	// in the old.

	std::vector<Window>::iterator iter;

	for (iter = old_list.begin(); iter != old_list.end(); iter++) {
		if (!contains_window(new_list, *iter)) {
			// the old one has this window but the new one does not. it was probably
			// closed.
			WindowDeltaPair pair;
			pair.added = false;
			pair.window = *iter;
			dest.push_back(pair);
		}
	}

	// getting here means that there have been no windows closed.
	for (iter = new_list.begin(); iter != new_list.end(); iter++) {
		if (!contains_window(old_list, *iter)) {
			// the new one has this window but the old one does not. it was probably
			// opened.
			WindowDeltaPair pair;
			pair.added = true;
			pair.window = *iter;
			dest.push_back(pair);
		}
	}
}

Visibility::Pager * pager = 0;

static void reload_config(int signal) {
	Visibility::Config * config = Visibility::Config::instance();
	config->reload();
	config->font_setup();
	config->signal_changed.emit();
}

std::map<Window, std::tr1::shared_ptr<Visibility::Window> > Visibility::Window::windows;
Visibility::Tooltip Visibility::WindowIcon::tooltip;

bool show_tooltip_wrap(void) {
	Visibility::WindowIcon::tooltip.map();
	Visibility::Server::instance()->flush();
	return false;
}

bool handle_x_event(int fd_x) {
	static std::vector<Window> client_list;
	static Visibility::Server * server = Visibility::Server::instance();
	static Visibility::Config * config = Visibility::Config::instance();
	static std::tr1::shared_ptr<MainLoop::FunctorBase> tooltip_timeout;

	if (!client_list.size()) server->get_client_list(client_list);

	while(XPending(server->display)) {
		XEvent event;
		XNextEvent(server->display, &event);
		switch (event.type) {
			case ButtonRelease:
				DEBUG_P("button release\n");
				if (event.xbutton.button != 1) break;
				Visibility::Window::get_window(event.xbutton.window)->signal_clicked.emit(event.xbutton.window);
				break;
			case EnterNotify:
				DEBUG_P("enter notify\n");
				Visibility::Window::get_window(event.xcrossing.window)->signal_pointer_entered.emit(event.xcrossing.window);
				if (!tooltip_timeout) {
					tooltip_timeout = MainLoop::ptr_fun(&show_tooltip_wrap);
					MainLoop::MainLoop::add_timeout_function(tooltip_timeout,
						(unsigned)(config->tooltip_time * 1e6));
				}
				break;
			case LeaveNotify:
				DEBUG_P("leave notify\n");
				Visibility::Window::get_window(event.xcrossing.window)->signal_pointer_left.emit(event.xcrossing.window);
				tooltip_timeout->disconnect();
				tooltip_timeout.reset();
				Visibility::WindowIcon::tooltip.unmap();
				break;
			case PropertyNotify:
				if (event.xproperty.atom == server->net_client_list_atom) {
					DEBUG_P("client list changed\n");
					// client list changed! onoes
					std::vector<Window> new_client_list;
					server->get_client_list(new_client_list);
					std::vector<WindowDeltaPair> changed;
					client_list_delta(client_list, new_client_list, changed);
					std::vector<WindowDeltaPair>::iterator iter;
					for (iter = changed.begin(); iter != changed.end(); iter++) {
						if (iter->added) {
							Visibility::Window::signal_new.emit(iter->window);
						} else {
							Visibility::Window::get_window(iter->window)->signal_closed.emit(iter->window);
						}
					}
					client_list = new_client_list;
				} else if (event.xproperty.atom == server->net_wm_icon_atom) {
					DEBUG_P("window icon changed\n");
					// a window's icon has changed 0:
					Visibility::Window::get_window(event.xproperty.window)->signal_icon_changed.emit(event.xproperty.window);
				} else if (event.xproperty.atom == server->wm_hints_atom) {
					DEBUG_P("XWMHints changed\n");
					// xwmhints contain window icon info as well as state info
					Visibility::Window::get_window(event.xproperty.window)->signal_icon_changed.emit(event.xproperty.window);
					Visibility::Window::get_window(event.xproperty.window)->signal_state_changed.emit(event.xproperty.window);
				} else if (event.xproperty.atom == server->net_wm_desktop_atom) {
					DEBUG_P("window changed desktop\n");
					if (event.xproperty.state == PropertyDelete) {
						// XXX: What should i really be doing here?
						Visibility::Window::get_window(event.xproperty.window)->select_input(NoEventMask);
					} else {
						// a window moved
						Visibility::Window::signal_window_desktop_changed.emit(event.xproperty.window);
					}
				} else if (event.xproperty.atom == server->net_current_desktop_atom) {
					// desktop swtich occurred
					DEBUG_P("current desktop changed\n");
					Visibility::Server::signal_desktop_changed.emit();
				} else if (event.xproperty.atom == server->net_wm_state_atom) {// || event.xproperty.atom == server->net_wm_window_type_atom) {
					DEBUG_P("window state changed\n");
					if (event.xproperty.state == PropertyDelete) {
						// XXX: What should i really be doing here?
						Visibility::Window::get_window(event.xproperty.window)->select_input(NoEventMask);
					} else {
						Visibility::Window::get_window(event.xproperty.window)->signal_state_changed.emit(event.xproperty.window);
					}
				} else if (event.xproperty.atom == server->net_wm_window_type_atom) {
					Visibility::Window::get_window(event.xproperty.window)->signal_type_changed.emit(event.xproperty.window);

				} else if (event.xproperty.atom == server->net_active_window_atom) {
					DEBUG_P("active window changed\n");
					Visibility::Window::signal_active_window_changed.emit();
				} else if (event.xproperty.atom == server->net_desktop_names_atom) {
					DEBUG_P("desktop names changed\n");
					Visibility::Server::signal_desktop_names_changed.emit();
				} else if (event.xproperty.atom == server->net_number_of_desktops_atom) {
					DEBUG_P("number of desktops changed\n");
					Visibility::Server::signal_number_of_desktops_changed.emit();
				}
		}
	}

	return true;
}

int main(int argc, char ** argv) {
	ArgParser parser;
	parser.register_option("config", "Config file to use. If the path does not start with '/', the\n\t\tfile will be looked for in ~/.config/visibility/.", true);

	if (!parser.parse(argc, argv)) {
		std::cerr << "Error parsing command line: " << parser.get_error() << std::endl;
		return 1;
	}

	if (parser.has_argument("help")) {
		std::cout << parser.help_text() << std::endl;
		return 0;
	}

	Visibility::Config * config = Visibility::Config::instance();

	if (parser.has_argument("config")) {
		config->set_config_file(parser.get_value("config"));
	}

	Visibility::Server * server = Visibility::Server::instance();
	CARD32 number_of_desktops = 0;

	XSetErrorHandler(x_error_handler);

	/*Visibility::Drawable::draw = XftDrawCreate(server->display, server->root_window, server->visual, server->colormap);*/
	config->font_setup();

	// wait until a window manager is started
	while ((number_of_desktops = server->get_number_of_desktops()) == 0) {
		sleep (1);
	}

	Visibility::Desktop::update_desktop_names();
	Visibility::Server::signal_desktop_names_changed.connect(sigc::ptr_fun(&Visibility::Desktop::update_desktop_names));
	Visibility::Server::signal_number_of_desktops_changed.connect(sigc::ptr_fun(&Visibility::Desktop::update_desktop_names));

	pager = new Visibility::Pager;

	XSelectInput(server->display, server->root_window, PropertyChangeMask);
	XFlush(server->display);

	signal(SIGHUP, reload_config);

	

	MainLoop::MainLoop::add_fd_watch(ConnectionNumber(server->display), MainLoop::ptr_fun(&handle_x_event));

	MainLoop::MainLoop::run();

	/*XftDrawDestroy(Visibility::Drawable::draw);*/

	return 0;
}
